from http import HTTPStatus
from urllib.parse import urljoin

from chalice import Blueprint, Response
from chalice.app import Request, ForbiddenError
from octokit import webhook
from pubsub import pub
from pubsub.utils import ExcPublisher

from zeroae.goblet.utils import (
    create_app_manifest,
    infer_app_url,
    get_create_app_url,
    get_configured_octokit,
)
from .. import config
from ..views import render_setup_html


class GitHubAppBlueprint(Blueprint):
    def __init__(self, import_name):
        """Creates the Chalice GitHubApp Blueprint."""
        super().__init__(import_name)
        pub.setListenerExcHandler(ExcPublisher(pub.getDefaultTopicMgr()))

    @staticmethod
    def on_gh_event(topic):
        """
        :param topic: The GitHub webhook topic
        :return: decorator
        """

        def decorator(f):
            l, _ = pub.subscribe(f, "gh" if topic == "*" else f"gh.{topic}")
            f.listener = l
            return f

        return decorator


bp: GitHubAppBlueprint = GitHubAppBlueprint(__name__)


@bp.route("/")
def register():
    """
    Note: HTTP Post Redirect, 307 code
    ref: bit.ly/why-doesnt-http-have-post-redirect
    :return:
    """
    # Get GitHub's create_app_url, for user/owner application
    create_app_url = get_create_app_url()

    request: Request = bp.current_request
    app_url = infer_app_url(request.headers, request.context["path"])
    app_manifest = create_app_manifest(app_url)
    body = render_setup_html(app_manifest, create_app_url)
    return Response(
        body=body, status_code=HTTPStatus.OK, headers={"Content-Type": "text/html"}
    )


@bp.route("/callback")
def register_callback():
    """
    Finish the GitHub application registration flow.

    Converts code for clientId, clientSecret, webhook secret, and App PEM
    Saves result in the configuration backend
    :return:
    """
    query_params = bp.current_request.query_params
    query_params = query_params if query_params else dict()

    code = query_params.get("code", None)
    if code is None:
        return Response(
            body='{"error": "The GitHub application-manifest code is missing."}',
            status_code=HTTPStatus.EXPECTATION_FAILED,
            headers={"Content-Type": "text/html"},
        )

    o = get_configured_octokit()
    o = o.apps.create_from_manifest(code=code)

    config.save_app_registration(o.json)

    return Response(
        body="",
        status_code=HTTPStatus.SEE_OTHER,
        headers={"Location": urljoin(f'{o.json["html_url"]}/', "installations/new")},
    )


@bp.route("/events", methods=["POST"])
def events():
    if not webhook.verify(
        bp.current_request.headers,
        bp.current_request.raw_body.decode("utf-8"),
        config.APP_WEBHOOK_SECRET,
        events=["*"],
    ):
        raise ForbiddenError(
            f"Error validating the event: {bp.current_request.to_dict()}"
        )

    r: Request = bp.current_request
    event_topic = r.headers["x-github-event"]
    pub.sendMessage(f"gh.{event_topic}", payload=r.json_body)
    return {}
